@@ -1,29 +1,44 @@
+import random
 import random as rd
 
 import numpy as np
 from tqdm import tqdm
 
+color1 = '\033[92m'
+color2 = '\033[94m'
+color3 = '\033[0m'
+
+print(
+    f"{color1}You use the library SimplePyQLearning !\n{color2}Credit: LeLaboDuGame on Twitch -> "
+    f"https://twitch.tv/LeLaboDuGame{color3}")
+
 
 class QLearning:
     """
-
-    SimplePyQLearning:
-    ----------------
-    Create by LeLaboDuGame
+    SIMPLEPYQLEARNING
+    -----------------
+    Hey guy !
+    I'm 15 years frensh dev !
+    I present you my project of q-qlearning !
+    Just for the credit : LeLaboDuGame on Twitch -> https://twitch.tv/LeLaboDuGame
+    You can use this library on all your project !
 
     Exemple:
     -------
-        map = [[0, -1, 0, 0, 1],
-        [0, 0, 0, -1, 0]]
+        map = [[0, -1, 0,  0, 1],
+       [0, -1, 0, -1, 0],
+       [0, -1, 0, -1, 0],
+       [0, -1, 0, -1, 0],
+       [0,  0, 0, -1, 0]]
         q = QLearning(action=[[-1, 0], [1, 0], [0, -1], [0, 1]], map=map, terminal_state=[-1, 1], epsylone=1,
-        epsylone_dec=0.0001, epsylone_min=0.1,
-        learning_rate=0.8, discount_factor=1, spawn=[0, 0])
-        q.train(1000)
-        print(q.get_all_cheminement([0, 0]))
+              epsylone_dec=0.0001, epsylone_min=0.1,
+              learning_rate=0.8, discount_factor=0.8, spawn="rnd")
+        q.train(10000)
+        print(q.get_all_cheminement([0, 0], timeout=1000, debug=False))
     """
 
-    def __init__(self, action, map, terminal_state, epsylone=0.7, epsylone_dec=0.01, epsylone_min=0.1,
-                 learning_rate=0.8, discount_factor=0.5, spawn=[]):
+    def __init__(self, action, map, terminal_state, epsylone=0.7, epsylone_dec=0.001, epsylone_min=0.1,
+                 learning_rate=0.8, discount_factor=0.8, spawn="rnd"):
         self.map = np.array(map)
         self.epsylone_min = epsylone_min
         self.epsylone_dec = epsylone_dec
@@ -31,11 +46,12 @@ class QLearning:
         self.discount_factor = discount_factor
         self.terminal_state = terminal_state
         self.spawn = spawn
-        if self.spawn == []:
-            for i in self.map.shape:
-                self.spawn.append(i)
+        if self.spawn == "random" or self.spawn == "rnd":
+            self.spawn = self.random_spawn()
 
-        self.spawn = (np.array(self.spawn), spawn == [])
+        print(f"Initiale spawn is {self.spawn}.")
+
+        self.spawn = (np.array(self.spawn), spawn == "rnd")
 
         self.epsylone = epsylone
         self.action = np.array(action)
@@ -51,6 +67,15 @@ class QLearning:
     def get_next_state(self, s, a):
         return s - self.action[a]
 
+    def random_spawn(self):
+        while True:
+            spawn = []
+            for i in self.map.shape:
+                spawn.append(random.randint(0, i - 1))
+            if not self.is_terminal_state(spawn):
+                break
+        return spawn
+
     def is_terminal_state(self, s):
         for i in self.terminal_state:
             try:
@@ -60,22 +85,30 @@ class QLearning:
                 return False
         return False
 
-    def get_all_cheminement(self, spawn):
+    def get_all_cheminement(self, spawn, timeout=10000, debug=True):
         s = spawn
         actions = []
         states = [s]
+        i = 0
         while not self.is_terminal_state(s):
-            if np.argmax(self.q_value[tuple(s)]) == 0:
-                while True:
-                    a = np.array(rd.randint(0, self.action.shape[0] - 1))
-                    if self.action_is_possible(s, a) and not self.is_terminal_state(self.get_next_state(s, a)):
-                        break
-            else:
-                a = np.argmax(self.q_value[tuple(s)])
+            if i == timeout:
+                print(
+                    f"{color1}Stoped because to much repetition.\nMaybe it's because it has a probleme somewhere. Verify the terminal_state values.{color3}")
+                timeout = True
+                break
+            x = 1
+            while True:
+                a = np.argsort(self.q_value[tuple(s)])[-x]
+                if self.action_is_possible(s, a):
+                    break
+                x += 1
             actions.append((a.tolist(), self.action[a].tolist()))
             s = self.get_next_state(s, a)
             states.append(s.tolist())
-
+            i += 1
+        if timeout == True:
+            if not debug:
+                return None
         print(self.show_player_pass(states))
         return actions, states, self.map[tuple(s)]
 
@@ -98,20 +131,25 @@ class QLearning:
     def train(self, episode):
 
         for i in tqdm(range(episode)):
-            s = self.spawn[0]
             if self.spawn[1]:
-                print(f"Repetition n°{i} at spawn : {s}")
+                s = self.random_spawn()
+            else:
+                s = self.spawn[0]
             while not self.is_terminal_state(s):
+                rnd = rd.random()
+                x = 1
                 while True:
                     if np.zeros((len(self.action))).tolist() == self.q_value[tuple(s)].tolist():
                         a = np.array(rd.randint(0, self.action.shape[0] - 1))
                     else:
-                        if rd.random() > self.epsylone:
-                            a = np.argmax(self.q_value[tuple(s)])
+                        if rnd > self.epsylone:
+                            a = np.argsort(self.q_value[tuple(s)])[-x]
                         else:
                             a = np.array(rd.randint(0, self.action.shape[0] - 1))
                     if self.action_is_possible(s, a):
                         break
+                    x += 1
+
                 old_s = s
                 s = self.get_next_state(s, a)
                 r = self.map[tuple(s)]
@@ -122,7 +160,7 @@ class QLearning:
                 # print(old_s, self.learning_rate, temporal_difference, new_q_value)
                 self.q_value[tuple(old_s)][a] = new_q_value
 
-            if self.epsylone > self.epsylone_min and episode % 10:
+            if self.epsylone > self.epsylone_min and episode % 10 == 0:
                 self.epsylone -= self.epsylone_dec
         print(self.q_value, self.epsylone)
         return self.q_value
